#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import configparser

# External imports
import sys
from argparse import ArgumentParser
from os import environ, mkdir, system
from os.path import abspath, dirname, exists, join
from shutil import copyfile, rmtree
from venv import create  # type: ignore

from dbgen import __file__ as DBGEN_ROOT

"""
Initialize a dbgen model

PROBLEMS

-path to .env and /data/ folders should be able to be specified (or by default be
    parallel )
- don't use DBGEN_ROOT -- use '__file__' instead to get current dbgen repo
"""
################################################################################

# Check to confirm that python3.6 or newer is being used
major_version, minor_version = sys.version_info[:2]
if major_version < 3 or minor_version < 6:
    raise Exception("Python 3.6+ is required.")

root = abspath(join(dirname(DBGEN_ROOT), ".."))
user = environ["USER"]
################################################################################


class File:
    def __init__(self, pth: str, template: str, fill_content: bool = True) -> None:
        self.pth = pth
        self.template = template
        self.fill_content = fill_content

    def content(self, kwargs: dict) -> str:
        with open(join(root, "dbgen/CLI/newfiles", self.template)) as f:
            if self.fill_content:
                return f.read().format(**kwargs)
            return f.read()

    def write(self, pth: str, **kwargs: str) -> None:
        with open(join(pth, self.pth), "w") as f:
            f.write(self.content(kwargs))


sch = File("schema.py", "schema.py")
ginit = File("generators/__init__.py", "ginit.py")
default = File("dbgen_files/default.py", "default.py")
io = File("generators/io.py", "io.py")
ana = File("generators/analysis.py", "analysis.py", fill_content=False)
man = File("main.py", "main.py")
data = File("data/example.csv", "data.csv")
parse = File("scripts/io/parse_employees.py", "parse.py")
utils = File("utils.py", "utils.py")
dev, log = [File(f"dbgen_files/{x}.json", x) for x in ["dev", "log"]]

files = [sch, ginit, default, man, io, data, parse, dev, log, utils, ana]
inits = ["", "scripts/", "scripts/io/"]
dirs = [
    "generators",
    "scripts",
    "data",
    "dbgen_files",
    "scripts/io",
    "dbgen_files/storage",
    "dbgen_files/tmp",
]


################################################################################
parser = ArgumentParser(description="Initialize a dbGen model", allow_abbrev=True)
parser.add_argument("--pth", type=str, help="Root folder", required=True)
parser.add_argument("--name", type=str, help="Name of model", required=True)
parser.add_argument("--env", default=".env/bin/activate", type=str, help="Name of model")
parser.add_argument("--create-env", action="store_true", help="Create a virtual env")

################################################################################
def create_config(model_name: str, model_root: str):
    model_root = abspath(model_root)
    envvars = dict(
        MODEL_NAME=model_name,
        MODEL_ROOT=model_root,
        MODEL_TMP=f"{model_root}/dbgen_files/tmp",
    )
    config = configparser.ConfigParser()
    config["dbgen"] = envvars
    with open(join(model_root, "model.cfg"), "w") as f:
        config.write(f)


################################################################################


def main(pth: str, name: str, env: str, create_env: bool = True) -> None:
    """
    Initialize a DbGen model
    #"""
    if exists(pth):
        print(pth, " already exists")
        while True:
            answer = input("Overwrite? (y/n)").lower()
            if answer == "y":
                rmtree(pth)
                break
            elif answer == "n":
                return
            else:
                print("invalid response")

    mkdir(pth)

    for dir in dirs:
        mkdir(join(pth, dir))
    for i in inits:
        system("touch " + join(pth, i + "__init__.py"))
    for fi in files:
        fi.write(pth, model=name, user=user)

    # Create virtual environment
    env = join(pth, env)
    reqs = join(root, "requirements.txt")
    if create_env:
        create(join(pth, ".env"), with_pip=True, symlinks=True, clear=True)
        system("source " + env + "; pip install -r " + reqs)
        copyfile(reqs, join(pth, "requirements.txt"))
    create_config(name, pth)


if __name__ == "__main__":
    args = parser.parse_args()
    main(args.pth, args.name, args.env, args.create_env)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import re
from bdb import BdbQuit
from typing import TYPE_CHECKING
from typing import List as L

from tqdm import tqdm

from dbgen.core.gen import Generator

# Internal imports
from dbgen.core.misc import ConnectInfo as ConnI
from dbgen.core.misc import Test, onlyTest, xTest
from dbgen.core.schema import PathEQ
from dbgen.utils.lists import concat_map
from dbgen.utils.sql import Error, sqlexecute, sqlselect
from dbgen.utils.str_utils import levenshteinDistance

# Internal
if TYPE_CHECKING:
    from dbgen.core.model.model import Model

    Model

########################################################


def run(
    self: "Model",
    conn: ConnI,
    meta_conn: ConnI,
    nuke: bool = False,
    add: bool = False,
    retry: bool = False,
    only: L[str] = [],
    xclude: L[str] = [],
    start: str = None,
    until: str = None,
    serial: bool = False,
    bar: bool = True,
    clean: bool = False,
    skip_row_count: bool = False,
    batch: int = None,
    print_logo: bool = True,
) -> None:
    """
    This method is point of the model: to run and generate a database according
    to the model's specified rules.

    - conn/meta_conn: information to connect to database and logging database
    - nuke: By default, this is not used. If True, everything except generators
            tagged "no_nuke" are purged. Otherwise, give a space separated list
            of generator names/tags. If a generator is purged, then any
            tables it populates will be truncated. Any columns it populates will be set all
            to NULL. Any generators with inputs OR outputs that have any overlap with the outputs
            of a purged generator will be purged themselves.
    - add: needed if new entities/columns have been added to the model (but not yet in DB)
    - retry: ignore repeat checking
    - only: only run generators with these names (or these tags)
    - xclude: do not run generators with these names (or these tags)
    - start: start at the generator with this name
    - until: stop at the generator with this name
    - serial: force all Generators to be run without parallelization
    - bar: show progress bars
    - skip_row_count: Skip the select count(1) for all gens (good for large queries)
    - clean: 'cleans up' implementation detail columns (deleted) for
    presentation of the resulting database to others...at the cost of not being
    able to call model.run() without nuking again (unless an 'unclean' method is
    written, which is in principle possible)

    """
    # Run tests on pyblocks
    # ----------------------
    self.test_transforms()

    # # Setup logger and config
    # # --------------------
    run_logger = logging.getLogger("dbgen.run")
    # Print to-do list for the model
    # ---------------------------------------
    todo = self._todo()
    if todo:
        run_logger.warning(
            "the following attributes do not have any generator "
            "to populate them: \n\t-" + "\n\t-".join(sorted(todo))
        )

    # Validate input
    # ----------------
    assert conn != meta_conn, "Main DB cannot be in same schema as logging DB"
    startErr = 'Starting generator ("start") must be a Generator name'
    assert not start or start in self.gens, startErr
    xclude_ = set(xclude)
    only_ = set(only)
    for w in only_ | xclude_:
        self._validate_name(w)

    # # Make sure no existing cxns to database
    # # ---------------------------------------
    # conn.kill()
    # meta_conn.kill()

    # Make metatables
    # ----------------
    run_id = self._make_metatables(
        mconn=meta_conn,
        conn=conn,
        nuke=nuke,
        retry=retry,
        only=list(sorted(only_)),
        xclude=list(sorted(xclude_)),
        start=start,
        until=until,
        bar=bar,
    )

    # Clean up database
    # -----------------
    if nuke:
        self.make_schema(conn=conn, nuke=nuke, bar=bar)
    elif add:
        msg = """
#######################################################################
!!!WARNING!!!!
Add is an extremely experimental feature. Existing rows in
modified tables are not deleted. If you added ID info to a table then
you need to manually truncate that table (and cascade to linked tables)
or else those tables will contain rows with missing ID info in their PKs.

Add should really only be used to add attributes to existing tables
or add new empty tables. Adding identifying FKs from existing tables to new tables
is very dangerous and manual truncation will be necessary.

I hope you know what you are doing!!!
!!!WARNING!!!!
#######################################################################
        """
        run_logger.warning(msg)
        for ta in tqdm(
            self.objs.values(),
            desc="Adding new tables",
            leave=False,
            disable=not bar,
        ):
            for sqlexpr in ta.create():
                try:
                    sqlexecute(conn.connect(), sqlexpr)
                except Error as e:
                    # Error code for duplicate table
                    if e.pgcode == "42701":
                        run_logger.debug("dup")
                        pass
                    # Error code for when a relation doesn't exist on a table we
                    # are adding
                    elif re.match(r'column "\w+" of relation "\w+" does not exist', str(e)):
                        run_logger.debug(f"PGERROR: {e}")
                        pass
                    else:
                        raise Error(e)
        for v in tqdm(self.viewlist, desc="Adding new views", leave=False, disable=not bar):
            try:
                sqlexecute(conn.connect(), v.create())
            except Error as e:
                # Error code for duplicate table
                if "already exists" in str(e):
                    run_logger.debug("dup")
                    pass
                else:
                    raise Error(e)

        for ta in tqdm(
            self.objs.values(),
            desc="Adding new columns",
            leave=False,
            disable=not bar,
        ):
            for sqlexpr in self.add_cols(ta):
                try:
                    sqlexecute(conn.connect(), sqlexpr)
                except Error as e:
                    # Error code for duplicate column
                    if e.pgcode == "42701":
                        pass
                    else:
                        raise Error(e)

    # Make 'global' database connections (active throughout whole process)
    # ----------------------------------------------------------------------
    gcxn = conn.connect()
    gmcxn = meta_conn.connect()

    # Initialize variables
    # ---------------------
    not_run = []  # type: L[str] ### List of Rules that were not run
    err_tot = 0  # total # of failed generators
    start_flag = False if start else True
    until_flag = True
    start_test = Test(lambda _, __: start_flag, lambda _: 'excluded')
    until_test = Test(lambda _, __: until_flag, lambda _: 'excluded')
    testdict = {xTest: [xclude_], start_test: [None], until_test: [None]}

    universe = self._get_universe()

    def update_run_status(status: str) -> None:
        update_run_status = """UPDATE run SET status=%s WHERE run_id=%s"""
        sqlexecute(gmcxn, update_run_status, [status, run_id])

    # Set status to running
    update_run_status("running")
    with tqdm(total=len(self.gens), position=0, disable=not bar) as tq:
        for gen in self.ordered_gens():

            # Initialize Variables
            # ---------------------
            name = gen.name
            tq.set_description(name)
            run_logger.info(f"Running {gen.name}...")

            # Set flags
            # --------------------------------
            if name == start:
                start_flag = True

            # Run tests to see whether or not the Generator should be run
            if only:  # only trumps everything else, if it's defined
                only_result = onlyTest(gen, only_)
                xclude_result = xTest(gen, xclude_)
                run = (only_result is True) and (xclude_result is True)
                if only_result is not True:
                    gen.update_status(gmcxn, run_id, only_result)
                elif xclude_result is not True:
                    gen.update_status(gmcxn, run_id, xclude_result)
            else:
                run = True  # flag for passing all tests
                for test, args in testdict.items():
                    test_output = test(gen, *args)  # type: ignore
                    if test_output is not True:
                        not_run.append(name)
                        gen.update_status(gmcxn, run_id, test_output)
                        run = False
                        break
            if run:
                try:
                    err_tot += self._run_gen(
                        universe=universe,
                        gen=gen,
                        gmcxn=gmcxn,
                        gcxn=gcxn,
                        mconn_info=meta_conn,
                        conn_info=conn,
                        run_id=run_id,
                        retry=retry,
                        serial=serial,
                        bar=bar,
                        user_batch_size=batch,
                        skip_row_count=skip_row_count,
                    )
                except (
                    Exception,
                    KeyboardInterrupt,
                    SystemExit,
                    BdbQuit,
                ) as exc:
                    # If a critical error is hit that doesn't raise
                    # ExternalError() we need to clean up
                    # Update the run
                    update_run_status("failed")
                    # Update the Gen
                    error = str(exc) if str(exc) else repr(exc)
                    gen.update_status(gmcxn, run_id, "failed", err=error)
                    raise

            tq.update()

            # Set flags
            # ----------
            if name == until:
                until_flag = False

    end = """UPDATE run SET delta=EXTRACT(EPOCH FROM age(CURRENT_TIMESTAMP,starttime)),
                            status='completed',
                            errs = %s
             WHERE run_id=%s"""

    sqlexecute(gmcxn, end, [err_tot, run_id])
    self.check_paths(conn)

    if clean:
        for o in self.objs:
            for c in ["deleted"]:
                q = f"ALTER TABLE {o} DROP COLUMN {c}"
                sqlexecute(gcxn, q)

    gcxn.close()
    gmcxn.close()
    if bar:
        run_logger.info("\nFinished.\n\t" + (f"did not run {not_run}" if not_run else "Ran all Rules"))


def validate_name(self: "Model", w: str) -> None:
    """
    Checks to make sure name - in an argument of model.run() - is valid,
    If not, throws error and suggests alternatives
    """
    match = False
    close = []

    def t(u: Generator) -> L[str]:
        return [u.name] + u.tags

    for n in concat_map(t, list(self.gens.values())):
        d = levenshteinDistance(w, n)
        upW, upN = max(len(w), 5), max(len(n), 5)  # variables for safe indexing
        if d == 0:
            match = True
            break
        elif d < 5 or w[:upW] == n[:upN]:
            close.append(n)  # keep track of near-misses
    if not match:
        did_you = f"Did you mean {close}" if close else ""
        raise ValueError(f"No match found for {w}\n{did_you}")


def check_patheq(self: "Model", p: PathEQ, db: ConnI) -> None:
    """
    Check whether a given database enforces the a path equality specification
    """
    paths = list(p)
    ids = {n: o.id_str for n, o in self.objs.items()}
    p1, p2 = paths
    sels = p1.select(self), p2.select(self)
    joins = map("\n\t".join, (p1.joins(ids, self), p2.joins(ids, self)))
    start = p1.start()
    st_id = self[start].id_str

    q = """
        SELECT "{0}"."{1}",
               {2},
               {3}
        FROM {0} AS "{0}"
        {4}
        {5}
        """
    args = [start, st_id, *sels, *joins]
    query = q.format(*args)
    out = sqlselect(db.connect(), query)
    for id, a, b in out:
        if a != b:
            err = "Path Equality check FAILED for {} # {}" + "\n{} -> {}" * 2
            eargs = (start, id, p1, a, p2, b)
            raise ValueError(err.format(*eargs))



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging

# External imports
from typing import TYPE_CHECKING, Any
from typing import Dict as D
from typing import List as L
from typing import Tuple as T

if TYPE_CHECKING:
    from dbgen.core.misc import ConnectInfo as ConnI
    from dbgen.core.model.model import Model
    from dbgen.utils.sql import Connection as Conn

    Model
    ConnI
    Conn

from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults


class DummyGenOperator(BaseOperator):
    """A custom airflow operator for the generator object"""

    template_fields = ()
    ui_color = "#b3cde0"

    @apply_defaults
    def __init__(
        self,
        objs: D[str, T[L[str], L[int], L[int]]],
        gen_name: str,
        gen_hash: int,
        db_conn_id: str,
        mdb_conn_id: str,
        retry: bool = False,
        serial: bool = False,
        bar: bool = False,
        user_batch_size: int = None,
        **kwargs: dict,
    ) -> None:
        super().__init__(**kwargs)  # type: ignore
        # Initialize variables
        self.objs = objs
        self.gen_name = gen_name
        self.gen_hash = gen_hash
        self.retry = retry
        self.serial = serial
        self.bar = bar
        self.user_batch_size = user_batch_size
        self.db_conn_id = db_conn_id
        self.mdb_conn_id = mdb_conn_id
        self._logger = logging.getLogger(f"dbgen.run.{gen_name}")

    def execute(self, context: Any) -> None:

        # Get the db connections
        self._logger.info("Dummy Generator Executing...")
        self._logger.info(f"Running Generator {self.gen_name}")
        self._logger.info(f"{self.objs=}")
        self._logger.info(f"{self.gen_name=}")
        self._logger.info(f"{self.gen_hash=}")
        self._logger.info(f"{self.retry=}")
        self._logger.info(f"{self.serial=}")
        self._logger.info(f"{self.bar=}")
        self._logger.info(f"{self.user_batch_size=}")
        self._logger.info(f"{self.db_conn_id=}")
        self._logger.info(f"{self.mdb_conn_id=}")
        self._logger.info(f"Done!")



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from dbgen.core.expr.sqltypes import Boolean, Decimal, Text, Varchar
from dbgen.core.schema import Attr, Entity
from dbgen.core.schema import UserRel as Rel

####
sample = Entity("sample", attrs=[Attr("id", identifying=True)])
####
science = Entity(
    "scientist",
    attrs=[
        Attr("firstname", Varchar()),
        Attr("lastname", Varchar()),
        Attr("ssn", identifying=True, desc="social security"),
    ],
)
####
proced = Entity("procedures", attrs=[Attr("procedure_name", Varchar(), identifying=True)])

####
hist = Entity(
    "history",
    "mapping table",
    attrs=[
        Attr("step", identifying=True, desc="order of operations"),
        Attr("timestamp", Varchar()),
        Attr("step_divides_ssn", Boolean()),
    ],
    fks=[
        Rel("sample", identifying=True),
        Rel("expt_type", "procedures", identifying=True),
        Rel(
            "operator",
            "scientist",
        ),
    ],
)

####
hd = Entity(
    "history_detail",
    "RDF triplestore",
    attrs=[
        Attr("name", Varchar(), identifying=True),
        Attr("value", Text()),
        Attr("dtype", Varchar()),
    ],
    fks=[Rel("history", identifying=True)],
)
#####
elec = Entity(
    "electrode",
    "Either an anode or cathode",
    attrs=[Attr("composition", Varchar())],
    fks=[Rel("sample", identifying=True)],
)
#####

anode = Entity("anode", fks=[Rel("electrode", identifying=True)])
cathode = Entity("cathode", fks=[Rel("electrode", identifying=True)])

#####
fc = Entity(
    "fuel_cell",
    "Combination of a particular anode and cathode sample during an expt",
    attrs=[
        Attr("expt_id", identifying=True),
        Attr("electrolyte", Varchar()),
        Attr("capacity", Decimal()),
        Attr("timestamp", Varchar()),
        Attr("calc_anode", Boolean()),
    ],
    fks=[Rel("anode"), Rel("cathode")],
)
#####
ec = Entity(
    "electrode_composition",
    "mapping table",
    attrs=[Attr("frac", Decimal())],
    fks=[Rel("electrode"), Rel("element")],
)
#####
elem = Entity(
    "element",
    "Atomic elements",
    attrs=[
        Attr("atomic_number", identifying=True),
        Attr("symbol", Varchar()),
        Attr("name", Varchar()),
        Attr("mass", Decimal()),
    ],
)

all = [
    sample,
    science,
    proced,
    hist,
    hd,
    elec,
    anode,
    cathode,
    fc,
    elem,
    ec,
]



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# External modules
import logging
from functools import partial
from math import ceil
from multiprocessing import cpu_count, get_context
from time import time
from typing import TYPE_CHECKING, Any
from typing import Dict as D
from typing import List as L
from typing import Set as S
from typing import Tuple as T
from typing import Union as U

from tqdm import tqdm

from dbgen.core.funclike import PyBlock
from dbgen.core.load import Load
from dbgen.core.misc import ConnectInfo as ConnI
from dbgen.utils.exceptions import DBgenExternalError, DBgenInternalError, DBgenSkipException
from dbgen.utils.lists import broadcast
from dbgen.utils.numeric import safe_div
from dbgen.utils.sql import Connection as Conn
from dbgen.utils.sql import DictCursor, fast_load, mkUpdateCmd, sqlexecute, sqlselect
from dbgen.utils.str_utils import hash_

# Internal
if TYPE_CHECKING:
    from dbgen.core.gen import Generator
    from dbgen.core.model.model import UNIVERSE_TYPE, Model
###########################################


def run_gen(
    self: "Model",
    universe: "UNIVERSE_TYPE",
    gen: "Generator",
    gmcxn: Conn,
    gcxn: Conn,
    mconn_info: ConnI,
    conn_info: ConnI,
    run_id: int,
    retry: bool = False,
    serial: bool = False,
    bar: bool = False,
    user_batch_size: int = None,
    skip_row_count: bool = False,
    gen_hash: str = None,
) -> int:
    """
    Executes a SQL query, then maps each output over a processing function.
    How the DB is modified is determined self.loads (list of InsertUpdate).
    """
    # Initialize Variables
    # --------------------
    logger = logging.getLogger(f"dbgen.run.{gen.name}")

    gen.update_status(gmcxn, run_id, "running")
    start = time()
    retry_ = retry or ("io" in gen.tags)
    if gen_hash:
        a_id = gen_hash
    else:
        a_id = gen.hash
    keys_to_save = gen._get_all_saved_key_dict()
    bargs = dict(leave=False, position=1, disable=not bar)
    bargs_inner = dict(leave=False, position=2, disable=not bar)
    parallel = (not serial) and ("parallel" in gen.tags)

    # Set the hasher for checking repeats
    ghash = gen.hash

    def hasher(x: Any) -> str:
        """Unique hash function to this Generator"""
        return hash_(str(ghash) + str(x))

    # Determine how to map over input rows
    # -------------------------------------
    cxns = (gmcxn, gcxn)

    cpus = cpu_count() - 1 or 1  # play safe, leave one free
    ctx = get_context(
        "forkserver"
    )  # addresses problem due to parallelization of numpy not playing with multiprocessing
    mapper = partial(ctx.Pool(cpus).imap_unordered, chunksize=5) if parallel else map

    # Wrap everything in try loop to catch errors
    try:
        # First setup query, get num_inputs and set batch size
        with tqdm(total=1, desc="Initializing Query", **bargs) as tq:
            logger.debug("Initializing Query")
            # Name the cursor for server side processing, need to turn off auto_commit
            cursor = conn_info.connect(auto_commit=False).cursor(
                f"{run_id}-{a_id}", cursor_factory=DictCursor
            )

            # If there is a query get the row count and execute it
            if gen.query:
                tq.set_description("Getting row count...")
                logger.info("Getting row count...")
                if not skip_row_count:
                    num_inputs = gen.query.get_row_count(gcxn)
                else:
                    num_inputs = int(10e8)
                logger.info(f"Number of inputs = {num_inputs}")
                logger.info("Executing query...")
                tq.set_description("Executing query...")
                cursor.execute(gen.query.showQ())
            # No query gens have 1 input
            else:
                num_inputs = 1

            # If user supplies a runtime batch_size it is used
            if user_batch_size is not None:
                batch_size: int = user_batch_size
                logger.info(f"Using user defined batch size: {user_batch_size}")
            # If generator has the batch_size set then that will be used next
            elif gen.batch_size is not None:
                batch_size = gen.batch_size
                logger.info(f"Using the {gen.name} specified batch size: {gen.batch_size}")
            elif skip_row_count:
                batch_size = 1
                logger.info(f"Using a default batch size of 1 since --skip-row-count flag is set")
            # Finally if nothing is the default is set to batchify the inputs into
            # 20 batches
            else:
                batch_size = ceil(num_inputs / 20) if num_inputs > 0 else 1
                logger.info(f"Using the default batch size to get 20 batches: {batch_size}")

            tq.update()

        logger.info("Applying...")
        # Wrap batch processing in try loop to close the curosr on errors
        try:
            # Iterate over the batches
            # Get the already processed input hashes from the metadb
            rpt_select = "SELECT repeats_id FROM repeats WHERE repeats.gen = %s"
            rpt_select_output = sqlselect(gmcxn, rpt_select, [a_id])
            rpt_select_output = [str(x[0]) for x in rpt_select_output]  # type: ignore
            rpt_select_output = set(rpt_select_output)  # type: ignore
            for _ in tqdm(range(ceil(num_inputs / batch_size)), desc="Applying", **bargs):
                # fetch the current batch of inputs
                if gen.query:
                    logger.debug(f"Fetching batch")
                    inputs = cursor.fetchmany(batch_size)
                    if inputs is None:
                        break
                else:
                    # if there is no query set the inputs to be length 1 with
                    # empty dict as input
                    inputs = [{}]
                # If retry is true don't check for repeats
                if retry_:
                    inputs = [(x, hasher(x)) for x in inputs]
                # Check for repeats
                elif inputs:
                    with tqdm(total=1, desc="Repeat Checking", **bargs_inner) as tq:
                        logger.debug(f"Repeat Checking")
                        # pair each input row with its hash
                        unfiltered_inputs = [(x, hasher(x)) for x in inputs]
                        # remove the hashes that are already in the processed_hashes
                        inputs = [(x, hx) for x, hx in unfiltered_inputs if hx not in rpt_select_output]
                        tq.update()

                # If we have no query or if we have any inputs we apply the ETL to the batch
                if gen.query is None or len(inputs) > 0:
                    apply_batch(
                        inputs=inputs,
                        f=gen.transforms,
                        acts=gen.loads,
                        universe=universe,
                        a_id=a_id,
                        qhsh=gen.query.hash if gen.query else "0",
                        run_id=run_id,
                        parallel=parallel,
                        mapper=mapper,
                        keys_to_save=keys_to_save,
                        cxns=cxns,
                        gen_name=gen.name,
                        bar=bar,
                    )
        finally:
            cursor.close()

        # Closing business
        # -----------------
        gen.update_status(gmcxn, run_id, "completed")
        tot_time = time() - start
        q = mkUpdateCmd("gens", ["runtime", "rate", "n_inputs"], ["run", "name"])
        runtime = round(tot_time / 60, 4)
        rate = round(safe_div(tot_time, num_inputs), 4)
        sqlexecute(gmcxn, q, [runtime, rate, num_inputs, run_id, gen.name])
        return 0  # don't change error count

    except DBgenExternalError as e:
        msg = f"\n\nError when running generator {gen.name}\n"
        logger.error(msg)
        logger.error(f"\n{e}")
        q = mkUpdateCmd("gens", ["error", "status"], ["run", "name"])
        sqlexecute(gmcxn, q, [str(e), "failed", run_id, gen.name])
        return 1  # increment error count


def transform_func(
    input: T[dict, str], qhsh: str, pbs: L[PyBlock], logger: logging.Logger
) -> U[T[dict, str], T[None, None]]:
    row, in_hash = input
    try:
        d = {qhsh: row}
        for pb in pbs:
            logger.debug(f"running {pb.func.name}")
            d[pb.hash] = pb(d)
    except DBgenSkipException as exc:
        logger.debug(f"Skipped {pb.func.name}: \nmsg:{exc.msg}")
        return None, None
    return d, in_hash


def delete_unused_keys(namespace: D[str, Any], keys_to_save: D[str, S[str]]) -> D[str, Any]:
    new_namespace = {}
    for hash_loc, names in keys_to_save.items():
        try:
            names_space_dict = namespace[hash_loc]
        except KeyError:
            raise DBgenInternalError(
                f"Looking for the namespace dict with keys: {names}, at location {hash_loc} but can't find it.\nDid you leave out the query?\nKeys:{list(namespace.keys())}"
            )
        pruned_dict = {key: val for key, val in names_space_dict.items() if key in names}
        new_namespace[hash_loc] = pruned_dict
    return new_namespace


def apply_batch(
    inputs: L[T[dict, str]],
    f: L[PyBlock],
    acts: L[Load],
    universe: "UNIVERSE_TYPE",
    a_id: str,
    run_id: int,
    qhsh: str,
    parallel: bool,
    mapper: Any,
    keys_to_save: D[str, S[str]],
    cxns: T[Conn, Conn],
    gen_name: str,
    bar,
) -> None:

    # Initialize variables
    logger = logging.getLogger(f"dbgen.run.{gen_name}.apply_batch")

    open_mdb, open_db = cxns
    n_inputs = len(inputs)
    n_loads = len(acts)
    processed_namespaces: L[D[str, Any]] = []
    processed_hashes: L[str] = []
    bargs = dict(leave=False, position=2, disable=not bar)

    logger.info("Transforming...")
    # Transform the data
    with tqdm(total=n_inputs, desc="Transforming", **bargs) as tq:
        transform_func_curr = partial(transform_func, pbs=f, qhsh=qhsh, logger=logger)
        for i, (output_dict, output_hash) in enumerate(mapper(transform_func_curr, inputs)):
            if output_dict:
                processed_namespaces.append(delete_unused_keys(output_dict, keys_to_save))
                if output_hash is not None:
                    processed_hashes.append(output_hash)
            logger.debug(f"Transformed {i}/{n_inputs}")
            tq.update()

    logger.info("Loading...")
    # Load the data
    with tqdm(total=len(acts), desc="Loading", **bargs) as tq:
        for i, a in enumerate(acts):
            a.act(
                cxn=open_db,
                universe=universe,
                rows=processed_namespaces,
                gen_name=gen_name,
            )
            logger.debug(f"Loaded {i+1}/{n_loads}")
            tq.update()

    # Store the repeats
    logger.info("Storing Repeats")
    with tqdm(total=1, desc="Storing Repeats", **bargs) as tq:
        repeat_values: T[L[str], L[int], L[str]] = broadcast([a_id, run_id, processed_hashes])
        table_name = "repeats"
        col_names = ["gen", "run", "repeats_id"]
        obj_pk_name = "repeats_id"
        fast_load(open_mdb, repeat_values, table_name, col_names, obj_pk_name)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import TYPE_CHECKING, Any
from typing import Callable as C
from typing import List as L
from typing import Optional as Opt
from typing import Set as S

if TYPE_CHECKING:
    from hypothesis.strategies import SearchStrategy

    SearchStrategy
from dbgen.core.expr.expr import Expr
from dbgen.core.fromclause import Path
from dbgen.core.schema import AttrTup, RelTup
from dbgen.utils.lists import flatten

Fn = C[[Any], str]  # type shortcut

########################
class PathAttr(Expr):
    def __init__(self, path: Opt[Path], attr: AttrTup) -> None:
        assert attr
        self.path = path or Path(attr.obj)
        self.attr = attr

    def __str__(self) -> str:
        return f'"{self.path}"."{self.attr.name}"'

    def __repr__(self) -> str:
        return f"PathAttr<{self.path}.{self.attr.name}>"

    @classmethod
    def _strat(cls) -> "SearchStrategy":
        from hypothesis.strategies import builds, from_type, none

        return builds(cls, path=none(), attr=from_type(AttrTup))

    ####################
    # Abstract methods #
    ####################
    def attrs(self) -> L["PathAttr"]:
        return [self]

    def fields(self) -> list:
        """
        List of immediate substructures of the expression (not recursive)
        """
        return []

    def show(self, f: Fn) -> str:
        """Apply function recursively to fields."""
        return f(self)

    @property
    def name(self) -> str:
        return self.attr.name

    @property
    def obj(self) -> str:
        return self.attr.obj

    def allrels(self) -> S[RelTup]:
        stack = list(self.path.fks)
        rels = set()
        while stack:
            curr = stack.pop(0)
            if not isinstance(curr, list):
                rels.add(curr.tup())
            else:
                assert not stack  # only the last element should be a list
                stack = flatten(curr)
        return rels


################################################################################
def expr_attrs(expr: Expr) -> L["PathAttr"]:
    """Recursively search for any Path (Expr) mentions in the Expr."""
    out = [expr] if isinstance(expr, PathAttr) else []  # type: L['PathAttr']
    if hasattr(expr, "fields"):
        for field in expr.fields():
            out.extend(expr_attrs(field))

    return out



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# External Modules
from os.path import join

# Internal Modules
from dbgen import ConnectInfo, parser

from .schema import make_model  # type: ignore
from .utils import get_config

"""
Run a model
"""
################################################################################


def main(args: dict) -> None:
    """
    Run the model with no extensions from command line.
    """
    config = get_config()
    model_root = config["dbgen"]["model_root"]
    model_name = config["dbgen"]["model_name"]
    m = make_model(model_name)
    db = ConnectInfo.from_file(join(model_root, "dbgen_files/dev.json"))
    mdb = ConnectInfo.from_file(join(model_root, "dbgen_files/log.json"))

    m.run(db, mdb, **args)


if __name__ == "__main__":
    args = parser.parse_args()
    main(vars(args))



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Airflow Operator for the intitialization of a run"""
# External imports
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from dbgen.core.gen import Generator
    from dbgen.core.misc import ConnectInfo as ConnI
    from dbgen.core.model.model import Model
    from dbgen.utils.sql import Connection as Conn

    Model
    ConnI
    Conn

from airflow.hooks.postgres_hook import PostgresHook
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT

# Internal Imports
from dbgen.core.misc import ConnectInfo as ConnI


class RunOperator(BaseOperator):
    """A custom airflow operator for starting runs and validating the inputs"""

    template_fields = ()
    ui_color = "#b3cde0"

    @apply_defaults
    def __init__(
        self,
        run_id: int,
        db_conn_id: str,
        mdb_conn_id: str,
        retry: bool = False,
        serial: bool = False,
        bar: bool = False,
        user_batch_size: int = None,
        **kwargs: dict,
    ) -> None:
        super()
        # Initialize variables
        self.run_id = run_id
        self.retry = retry
        self.serial = serial
        self.bar = bar
        self.user_batch_size = user_batch_size
        self.db_conn_id = db_conn_id
        self.mdb_conn_id = mdb_conn_id

    def _get_run_variable(self, mcxn: "Conn") -> "Generator":
        raise NotImplementedError

    def execute(self, context: Any) -> None:

        # Get the db connections
        gcxn = PostgresHook(self.db_conn_id).get_conn()
        mgcxn = PostgresHook(self.mdb_conn_id).get_conn()

        gcxn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
        mgcxn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)

        # conn_info = ConnI.from_postgres_hook(PostgresHook.get_connection(self.db_conn_id))
        # mconn_info = ConnI.from_postgres_hook(PostgresHook.get_connection(self.mdb_conn_id))
